from django.core.exceptions import ValidationError


def mix(*cls):
    return type("", cls, {})


class MinMaxValidationMixin:
    min_value = None
    max_value = None

    def validate_value(self, value):
        value = super().validate_value(value)

        is_none_valid = self.min_value is None and self.max_value is None
        if not is_none_valid and value is None:
            raise ValidationError("Value cannot be None")

        if self.min_value is not None and value < self.min_value:
            raise ValidationError(f"Value cannot be less than {self.min_value}")

        if self.max_value is not None and value > self.max_value:
            raise ValidationError(f"Value cannot be greater than {self.max_value}")
        return value

    def get_help_format(self):
        yield from super().get_help_format()

        if self.min_value is not None:
            yield f" from {self.min_value}"

        if self.max_value is not None:
            yield f" to {self.max_value}"


class PositiveValidationMixin(MinMaxValidationMixin):
    min_value = 0



"""
General idea and general principle of caching.

* all prepared values are stored in local thread and pupulated from db during the launching the app
* before each request the system checks if the value was changed (since we assume that changing settings values is rare). In that case we know all of the settings are actual and synced from the very beginning and no possible changes in the middle of running.
* settings version is a guaratee that value can be parsed by the settings (the version can only be changed in code)
* the cache value consist of two hashes. Has of values and hash of versions
* the changing of cache value is a signal that the values should be reloaded from db
* the version part of the cache value is used to check if the value can be parsed by the settings. If versions doesn't match, it means that values should not be updated from db. Probably the service is deploying now.
* the change in admin doesn't update the local values, the local values will be updated only before the request
"""

from threading import local
import hashlib
from functools import lru_cache

from django.core.cache import caches

from .settings import CHECKSUM_KEY_PREFIX, CACHE_SPLITER, CACHE_TIMEOUT, CACHE_BACKEND


DATA = local()


def get_cache():
    return caches[CACHE_BACKEND]


@lru_cache(maxsize=None)
def get_cache_key():
    from .conf import ALL

    return calc_checksum({name: ALL[name].version for name in ALL.keys()})


def calc_checksum(values):
    """
    generate md5 hash for a dict with keys and values as strings
    """
    return hash_value(
        CACHE_SPLITER.join(values[name] for name in sorted(values.keys()))
    )


def hash_value(value):
    """
    generate md5 hash for a string
    """
    return hashlib.md5(value.encode("utf-8")).hexdigest()


def set_local_checksum():
    DATA.ALL_VALUES_CHECKSUM = calc_checksum(DATA.ALL_RAW_VALUES)


def push_checksum(value=None, key=None):
    if key is None:
        key = get_cache_key()

    if value is None:
        value = DATA.ALL_VALUES_CHECKSUM

    get_cache().set(CHECKSUM_KEY_PREFIX + key, value, CACHE_TIMEOUT)


def get_checksum_from_store():
    return get_cache().get(CHECKSUM_KEY_PREFIX + get_cache_key())


def get_checksum_from_local():
    return DATA.ALL_VALUES_CHECKSUM


def set_new_value(name, new_value, version=None):
    from .conf import ALL

    prev_value = DATA.ALL_RAW_VALUES.get(name)

    if version is not None and ALL[name].version == version and prev_value != new_value:
        DATA.ALL_RAW_VALUES[name] = new_value
        DATA.ALL_VALUES[name] = ALL[name].to_python(new_value)

    return prev_value


def get_value(name):
    assert DATA.POPULATED
    return DATA.ALL_VALUES[name]


def is_populated():  # better name?
    return getattr(DATA, "POPULATED", False)


def get_db_objects():
    from .models import ContentSetting

    return {v.name: v for v in ContentSetting.objects.all()}


def reset_all_values(trigger_checksum=None):
    if not is_populated():
        DATA.ALL_VALUES = {}
        DATA.ALL_RAW_VALUES = {}
        DATA.ALL_VALUES_CHECKSUM = None
        DATA.POPULATED = False

    from .conf import ALL

    try:
        db = get_db_objects()
    except Exception:
        DATA.POPULATED = False
        return

    DATA.POPULATED = True
    for name in ALL.keys():
        if name in db:
            set_new_value(name, db[name].value, version=db[name].version)
        else:
            set_new_value(name, ALL[name].default, ALL[name].version)

    assert len(DATA.ALL_VALUES) == len(ALL), "Not all values are populated"

    set_local_checksum()

    if trigger_checksum != get_checksum_from_local():
        push_checksum()


def check_update():
    if not is_populated():
        return

    store_checksum = get_checksum_from_store()
    if store_checksum is not None and store_checksum != get_checksum_from_local():
        reset_all_values(trigger_checksum=store_checksum)


def recalc_checksums():
    """ """

    from .conf import ALL

    db = get_db_objects()

    db_versions = {}
    db_values = {}

    for name in set(ALL.keys()) | set(db.keys()):
        if name in db:
            db_versions[name] = db[name].version
            db_values[name] = db[name].value
        else:
            db_versions[name] = ALL[name].version
            db_values[name] = ALL[name].default

    db_version_key = calc_checksum(db_versions)
    push_checksum(calc_checksum(db_values), db_version_key)

    if db_version_key == get_cache_key() or not is_populated():
        return

    local_values = {}
    for name in ALL.keys():
        if name in db:
            db_version = db[name].version
            if db_version != ALL[name].version:
                local_values[name] = DATA.ALL_RAW_VALUES[name]
            else:
                local_values[name] = db[name].value
        else:
            local_values[name] = ALL[name].default

    push_checksum(calc_checksum(local_values))



from contextlib import contextmanager, ContextDecorator

CONTEXT_STACK = []


@contextmanager
def context(**kwargs):
    CONTEXT_STACK.append(kwargs)
    try:
        yield
    finally:
        CONTEXT_STACK.pop()


def get_context_stack(name):
    for obj in CONTEXT_STACK:
        if name in obj:
            yield obj[name]


class content_settings_context(ContextDecorator):
    def __init__(self, **values) -> None:
        super().__init__()
        self.values_to_update = values
        self.prev_values = {}

    def __enter__(self):
        from content_settings.caching import set_new_value

        for name, new_value in self.values_to_update.items():
            self.prev_values[name] = set_new_value(name, new_value)

    def __exit__(self, *exc):
        from content_settings.caching import set_new_value

        for name, new_value in self.prev_values.items():
            set_new_value(name, new_value)



from collections import defaultdict

from django.db import models
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.conf import settings


def validate_content_setting_name(value):
    from .conf import ALL

    if value not in ALL:
        raise ValidationError(
            "Unknown code. Please verify, it should be one from content_settings.py"
        )


class ContentSetting(models.Model):
    name = models.CharField(
        max_length=200, unique=True, validators=[validate_content_setting_name]
    )
    value = models.TextField(blank=True)
    version = models.CharField(max_length=50, null=True)

    class Meta:
        ordering = ("name",)

    def __str__(self):
        return self.name


class HistoryContentSetting(models.Model):
    CHANGED_CHOICES = (
        (True, "Changed"),
        (False, "Added"),
        (None, "Removed"),
    )
    BY_USER_CHOICES = (
        (True, "by user"),
        (False, "by app"),
        (None, "unknown"),
    )

    created_on = models.DateTimeField(null=False, default=timezone.now)

    name = models.CharField(max_length=200)
    value = models.TextField(blank=True, null=True)
    version = models.CharField(max_length=50, null=True)

    was_changed = models.BooleanField(null=True, default=True, choices=CHANGED_CHOICES)
    by_user = models.BooleanField(null=True, default=None, choices=BY_USER_CHOICES)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, null=True
    )

    class Meta:
        ordering = ("-id",)
        indexes = [
            models.Index(fields=["name", "-id"]),
        ]

    @classmethod
    def update_last_record_for_name(cls, name, user=None):
        last_setting = cls.objects.filter(name=name).first()
        last_setting.user = user
        last_setting.by_user = user is not None
        last_setting.save()

    @classmethod
    def gen_unique_records(cls, name):
        next_obj = None
        records = cls.objects.filter(name=name)
        if not records.exists():
            return []

        for obj in records:
            if next_obj is None:
                next_obj = obj
                continue

            if next_obj.value != obj.value or next_obj.was_changed != obj.was_changed:
                yield next_obj

            next_obj = obj

        yield obj


# Model that hold user attached tags to content settings
class UserTagSetting(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    tag = models.TextField()

    class Meta:
        unique_together = (("user", "name", "tag"),)

    def __str__(self):
        return f"{self.user} - {self.name} - {self.tag}"

    @classmethod
    def get_user_settings(cls, user):
        settings = defaultdict(set)
        for name, tag in cls.objects.filter(user=user).values_list("name", "tag"):
            settings[name].add(tag)
        return settings



from functools import cached_property

from django.core.exceptions import ValidationError

from .basic import SimpleText, BaseSetting, SimpleInt


def f_empty(value):
    value = value.strip()
    if not value:
        return None
    return value


def f_comment(prefix):
    def _(value):
        if value.strip().startswith("#"):
            return None
        return value

    return _


class SimpleStringsList(SimpleText):
    comment_starts_with = "#"
    filter_empty = True
    split_lines = "\n"
    filters = None

    def get_filters(self):
        if self.filters is not None:
            return self.filters

        filters = []
        if self.filter_empty:
            filters.append(f_empty)

        if self.comment_starts_with:
            filters.append(f_comment(self.comment_starts_with))

        return filters

    def get_help_format(self):
        yield "List of values with the following format:"
        yield "<ul>"
        if self.split_lines == "\n":
            yield "<li>each line is a new value</li>"
        else:
            yield f"<li>{self.split_lines} separates values</li>"
        yield "<li>strip spaces from the beginning and from the end of the value</li>"
        yield "<li>remove empty values</li>"
        if self.comment_starts_with:
            yield f"<li> use {self.comment_starts_with} to comment a line</li>"
        if self.filter_empty:
            yield "<li>empty values are removed</li>"
        yield "</ul>"

    def filter_line(self, line):
        for f in self.get_filters():
            line = f(line)
            if line is None:
                break
        return line

    def gen_to_python(self, value):
        lines = value.split(self.split_lines)
        for line in lines:
            line = self.filter_line(line)

            if line is not None:
                yield line

    def to_python(self, value):
        return list(self.gen_to_python(value))


class TypedStringsList(SimpleStringsList):
    line_type = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        assert self.line_type is not None, "line_type should be set"
        assert isinstance(
            self.line_type, BaseSetting
        ), "line_type should be BaseSetting"

    def validate_value(self, value):
        lines = value.split(self.split_lines)
        for num, line in enumerate(lines, 1):
            line = self.filter_line(line)
            if line is None:
                continue

            try:
                self.line_type.validate_value(line)
            except ValidationError as e:
                raise ValidationError(f"Line {num}: {e.message}")

        return super().validate_value(value)

    def gen_to_python(self, value):
        for line in super().gen_to_python(value):
            yield self.line_type.to_python(line)

    def get_help_format(self):
        yield from super().get_help_format()
        yield "Each line is "
        yield from self.line_type.get_help_format()



import urllib.parse
from collections import defaultdict

from django.contrib import admin
from django.forms import ModelForm
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.contrib.admin.views.main import ChangeList
from django.urls import path
from django.http import JsonResponse
from django.contrib.admin.utils import unquote
from django.core.exceptions import PermissionDenied
from django.utils.text import capfirst
from django.template.response import TemplateResponse
from django.contrib.messages import add_message, ERROR
from django.utils.translation import gettext as _


from .models import ContentSetting, HistoryContentSetting, UserTagSetting
from .context_managers import content_settings_context
from .conf import ALL
from .settings import USER_TAGS


def user_able_to_update(user, name):
    return ALL[name].update_permission is None or ALL[name].update_permission(user)


TAGS_PARAM = "tags"
TAGS_SPLITER = "|"


def get_selected_tags_from_params(params):
    if TAGS_PARAM not in params:
        return set()
    return set(params[TAGS_PARAM].split(TAGS_SPLITER))


class SettingsChangeList(ChangeList):
    def get_filters_params(self, *args, **kwargs):
        params = super().get_filters_params(*args, **kwargs)
        if TAGS_PARAM in params:
            del params[TAGS_PARAM]
        return params

    def get_queryset(self, request, *args, **kwargs):
        q = super().get_queryset(request, *args, **kwargs)

        tags = get_selected_tags_from_params(self.params)
        if not tags:
            return q

        all_keys = []
        user_settings = UserTagSetting.get_user_settings(request.user)
        for name, val in ALL.items():
            val_tags = val.get_tags() | user_settings[name]
            if not val_tags or tags - val_tags:
                continue

            all_keys.append(name)

        return q.filter(name__in=all_keys)


class ContentSettinForm(ModelForm):
    def __init__(self, *args, **kwargs):
        super(ContentSettinForm, self).__init__(*args, **kwargs)
        self.fields["value"].strip = False

    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)

    class Meta:
        model = ContentSetting
        fields = ["value"]


class ContentSettingAdmin(admin.ModelAdmin):
    list_display = ["name", "value", "help", "tags", "marks"]
    list_editable = [
        "value",
    ]
    search_fields = ["name", "value"]
    readonly_fields = [
        "py_data",
        "name",
        "help",
        "version",
        "default_value",
        "fetch_one_setting",
    ]
    history_list_display = ["value"]
    form = ContentSettinForm

    def save_model(self, request, obj, form, change):
        if user_able_to_update(request.user, obj.name):
            super().save_model(request, obj, form, change)
            HistoryContentSetting.update_last_record_for_name(obj.name, request.user)
        else:
            add_message(
                request,
                ERROR,
                _("You are not allowed to change %(name)s") % {"name": obj.name},
            )

    def get_search_results(self, request, queryset, search_term):
        qs, search_use_distinct = super().get_search_results(
            request, queryset, search_term
        )

        if search_term:
            help_names = [
                name
                for name, value in ALL.items()
                if search_term.lower() in value.help.lower()
            ]
            help_names = set(help_names).difference(
                set(qs.values_list("name", flat=True))
            )
            if help_names:
                qs = qs.union(ContentSetting.objects.filter(name__in=help_names))

        return qs, search_use_distinct

    def get_changelist(self, request, **kwargs):
        return SettingsChangeList

    def context_tags(self, request):
        extra_context = {}
        selected_tags = get_selected_tags_from_params(request.GET)

        def q_tags(tags):
            if not tags:
                return "?"
            return "?" + TAGS_PARAM + "=" + urllib.parse.quote(TAGS_SPLITER.join(tags))

        def q_add_tag(tag):
            return q_tags(selected_tags | set([tag]))

        def q_remove_tag(tag):
            return q_tags(selected_tags - set([tag]))

        extra_context["selected_tags"] = [
            {
                "tag": USER_TAGS[tag][0] if tag in USER_TAGS else tag,
                "url": q_remove_tag(tag),
            }
            for tag in sorted(selected_tags)
        ]

        tags_stat = defaultdict(int)
        user_settings = UserTagSetting.get_user_settings(request.user)
        for name, val in ALL.items():
            val_tags = val.get_tags() | user_settings[name]
            if selected_tags and (not val_tags or selected_tags - val_tags):
                continue

            for tag in val_tags - selected_tags:
                tags_stat[tag] += 1

        extra_context["available_tags"] = [
            {
                "tag": USER_TAGS[tag][0] if tag in USER_TAGS else tag,
                "url": q_add_tag(tag),
                "total": total,
            }
            for tag, total in sorted(
                tags_stat.items(), key=lambda a: a[0] and a[0] not in USER_TAGS
            )
        ]

        return extra_context

    def changelist_view(self, request, extra_context=None):
        return super().changelist_view(
            request, {**(extra_context or {}), **self.context_tags(request)}
        )

    def context_tags_view(self, request):
        return TemplateResponse(
            request,
            "admin/content_settings/contentsetting/context_tags.html",
            self.context_tags(request),
        )

    def history_view(self, request, object_id, extra_context=None):

        # First check if the user can see this history.
        model = self.model
        obj = self.get_object(request, unquote(object_id))
        if obj is None:
            return self._get_obj_does_not_exist_redirect(
                request, model._meta, object_id
            )

        if not self.has_view_or_change_permission(request, obj):
            raise PermissionDenied

        opts = model._meta
        app_label = opts.app_label

        context = {
            **self.admin_site.each_context(request),
            "title": f"Change history: {obj}",
            "subtitle": None,
            "module_name": str(capfirst(opts.verbose_name_plural)),
            "object": obj,
            "opts": opts,
            "preserved_filters": self.get_preserved_filters(request),
            "action_list": HistoryContentSetting.gen_unique_records(obj.name),
            **(extra_context or {}),
        }

        request.current_app = self.admin_site.name

        return TemplateResponse(
            request, f"admin/{app_label}/{opts.model_name}/object_history.html", context
        )

    def get_changelist_form(self, request, **kwargs):
        return ContentSettinForm

    def has_delete_permission(self, *args):
        return False

    def help(self, obj):
        try:
            help = ALL[obj.name].get_help()
        except KeyError:
            return "(the setting is not using)"
        if obj.value.replace("\r", "") != ALL[obj.name].default:
            help = "<i>(value is not default)</i> <br>" + help
        return mark_safe(help)

    def tags(self, obj):
        try:
            tags = ALL[obj.name].get_tags()
        except KeyError:
            return "(the setting is not using)"
        else:
            if not tags:
                return "---"
            return ", ".join(tags)

    def marks(self, obj):
        return mark_safe(
            "".join(
                f"""<span>
            <span class="user_tag user_tag_selected user_tag_hidden" data-tag="{tag}" data-name="{obj.name}">{marks[0]}</span>
            <span class="user_tag user_tag_hidden" data-tag="{tag}" data-name="{obj.name}">{marks[1]}</span>
            </span>"""
                for tag, marks in USER_TAGS.items()
            )
        )

    def py_data(self, obj):
        try:
            return mark_safe(ALL[obj.name].get_admin_preview_value(obj.value, obj.name))
        except KeyError:
            return ""

    def default_value(self, obj):
        from django.utils.html import escape

        attr_default = (
            ALL[obj.name]
            .default.replace("&", "&amp;")
            .replace("\n", "&NewLine;")
            .replace('"', "&quot;")
        )
        try:
            return mark_safe(
                f"""
            <pre class="default_value">{escape(ALL[obj.name].default)}</pre>
            <a class="reset_default" data-value="{attr_default}" style="cursor: pointer; color:var(--link-fg)">Reset ("save" is required for applying)</a>
            """
            )
        except KeyError:
            return ""

    def fetch_one_setting(self, obj):
        if not obj.name:
            return ""

        def get_fetch_reverse(name):
            return reverse(
                "content_settings:fetch_one_setting",
                kwargs={"name": name.lower().replace("_", "-")},
            )

        urls = [get_fetch_reverse(obj.name)]
        fetch_groups = ALL[obj.name].fetch_groups
        if fetch_groups is not None:
            if isinstance(fetch_groups, str):
                fetch_groups = [fetch_groups]

            urls += [get_fetch_reverse(n) for n in fetch_groups]

        return "\n".join(urls)

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "preview/",
                self.admin_site.admin_view(self.preview_setting),
                name="preview_setting",
            ),
            path("add-tag/", self.admin_site.admin_view(self.add_tag), name="add_tag"),
            path(
                "remove-tag/",
                self.admin_site.admin_view(self.remove_tag),
                name="remove_tag",
            ),
            path(
                "context-tags/",
                self.admin_site.admin_view(self.context_tags_view),
                name="context_tags",
            ),
        ]
        return custom_urls + urls

    # ajax view that creates UserTagSetting
    def add_tag(self, request):
        try:
            name = request.POST["name"]
            tag = request.POST["tag"]
            UserTagSetting.objects.create(
                user=request.user,
                name=name,
                tag=tag,
            )
        except Exception as e:
            return JsonResponse(
                {
                    "error": str(e),
                }
            )
        else:
            return JsonResponse(
                {
                    "success": True,
                }
            )

    # ajax view that removes UserTagSetting
    def remove_tag(self, request):
        try:
            name = request.POST["name"]
            tag = request.POST["tag"]
            UserTagSetting.objects.get(
                user=request.user,
                name=name,
                tag=tag,
            ).delete()
        except Exception as e:
            return JsonResponse(
                {
                    "error": str(e),
                }
            )
        else:
            return JsonResponse(
                {
                    "success": True,
                }
            )

    def preview_setting(self, request):
        setting = ALL[request.POST["name"]]
        value = request.POST["value"]

        other_values = {
            name[2:]: value
            for name, value in request.POST.items()
            if name.startswith("o_")
        }
        try:
            setting.validate_value(value)
        except Exception as e:
            return JsonResponse(
                {
                    "error": str(e),
                }
            )
        with content_settings_context(**other_values):
            return JsonResponse(
                {
                    "html": setting.get_admin_preview_value(
                        value, request.POST["name"]
                    ),
                }
            )


admin.site.register(ContentSetting, ContentSettingAdmin)

"""
TODO:
* remove Add form
"""



# some job urls declared in api/urls.py
from django.urls import path
from .views import fetch_one_setting

app_name = "content_settings"
urlpatterns = [
    path("fetch/<str:name>/", fetch_one_setting, name="fetch_one_setting"),
]



from inspect import isclass
from pprint import pformat
from functools import reduce
from importlib import import_module

from django import forms

from content_settings.context_managers import get_context_stack
from content_settings.settings import CACHE_SPLITER, BASE_SETTINGS_CLASS_NAME


class BaseSetting:
    pass


module_path, class_name = BASE_SETTINGS_CLASS_NAME.rsplit(".", 1)
module = import_module(module_path)
BaseSetting = getattr(module, class_name)


class SimpleString(BaseSetting):
    lazy_function = None
    cls_field = forms.CharField
    widget = forms.TextInput
    widget_attrs = None
    fetch_permission = None
    update_permission = None  # TODO: permission for update/reset/delete setting
    help_format = None
    help = None
    template = False
    value_required = False
    version = ""
    tags = None
    fetch_groups = None
    validators = ()
    empty_is_none = False

    def __init__(self, default="", **kwargs):
        kwargs = self.update_kwargs_context(kwargs)
        for k, v in kwargs.items():
            if not hasattr(self, k):
                raise ValueError("Attribute {} not found".format(k))
            setattr(self, k, v)

        self.default = default
        self.field = self.get_field()

        assert isinstance(self.default, str), "Default should be str"
        assert isinstance(self.version, str), "Version should be str"
        assert (
            CACHE_SPLITER not in self.version
        ), f"Version should not contain CACHE_SPLITER:{CACHE_SPLITER}"

    def update_kwargs_context(self, kwargs):
        return {
            **kwargs,
            "tags": reduce(
                lambda a, b: a | set(b),
                get_context_stack("tags"),
                set(kwargs.get("tags", set())),
            ),
        }

    def get_help_format(self):
        return self.help_format

    def get_help(self):
        help_format = self.get_help_format()
        if help_format is not None and not isinstance(help_format, str):
            help_format = "".join(help_format)
        if help_format:
            return f"{self.help} <br><br> {help_format}"
        return self.help

    def get_tags(self):
        tags = self.tags
        if tags is None:
            tags = set()
        return set(tags)

    def get_field(self):
        return self.cls_field(
            widget=self.get_widget(),
            validators=(self.validate_value,),
            required=self.value_required,
        )

    def get_widget(self):
        if isclass(self.widget) and self.widget_attrs is not None:
            return self.widget(attrs=self.widget_attrs)
        else:
            return self.widget

    def validate_value(self, value):
        val = self.to_python(value)
        for validator in self.validators:
            validator(val)
        return val

    def to_python(self, value):
        if self.empty_is_none and value.strip() == "":
            return None
        return self.field.to_python(value)

    def json_view_value(self, request, value):
        return value

    def get_admin_preview_value(self, value, name):
        return "<pre>{}</pre>".format(
            pformat(
                self.to_python(value),
            )
        )


class SimpleText(SimpleString):
    widget = forms.Textarea
    widget_attrs = {"rows": 10, "cols": 80}


class URLString(SimpleString):
    cls_field = forms.URLField
    widget = forms.URLInput
    widget_attrs = {"style": "max-width: 600px; width: 100%"}


class SimpleInt(SimpleString):
    cls_field = forms.IntegerField

    def get_help_format(self):
        yield "Any number"


class SimpleBool(SimpleInt):
    min_value = 0
    max_value = 1
    empty_is_none = True

    def to_python(self, value):
        return bool(super().to_python(value))


class SimpleDecimal(SimpleString):
    cls_field = forms.DecimalField



from itertools import chain

from django.core.exceptions import ValidationError
from django.template import Context, Template

from .basic import SimpleString, SimpleText


class TemplateObject:
    template_obj_key = "obj"
    help_template_obj = "Main Object"
    test_obj_query = None
    template_static = None

    def get_help_format(self):
        yield "Simple <a href='https://docs.djangoproject.com/en/3.2/topics/templates/' target='_blank'>Django Template</a>. Available objects: <ul>"
        yield f"<li>{self.template_obj_key} - {self.help_template_obj}</li>"
        for name in chain(self.get_static_data().keys(), self.get_test_kwargs().keys()):
            yield f"<li>{name}</li>"
        yield "</ul>"

    def get_static_data(self):
        return self.template_static if self.template_static is not None else {}

    def to_python(self, value):
        template = Template(value.strip())

        def _(obj, **kwargs):
            return template.render(
                Context(
                    dict(
                        **{self.template_obj_key: obj},
                        **self.get_static_data(),
                        **kwargs,
                    )
                )
            )

        return _

    def get_test_obj(self):
        if self.test_obj_query:
            return self.test_obj_query.first()

    def get_test_kwargs(self):
        return {}

    def get_admin_preview_value(self, value, *args):
        obj = self.get_test_obj()
        kwargs = self.get_test_kwargs()

        if obj is not None:
            return self.to_python(value)(obj, **kwargs)

        return "None preview object (get_test_obj)"

    def validate_value(self, value):
        obj = self.get_test_obj()
        kwargs = self.get_test_kwargs()

        if obj is None:
            return

        try:
            return self.to_python(value)(obj, **kwargs)
        except Exception as e:
            raise ValidationError(str(e))


class StringTemplateObject(TemplateObject, SimpleString):
    widget_attrs = {"style": "max-width: 600px; width: 100%"}


class HTMLTemplateObject(TemplateObject, SimpleText):
    pass


class TextTemplateObject(HTMLTemplateObject):
    def get_admin_preview_value(self, value, *args):
        return super().get_admin_preview_value(value, *args).replace("\n", "<br>")
